/*----- PROTECTED REGION ID(AlarmNotify.h) ENABLED START -----*/
//=============================================================================
//
// file :        AlarmNotify.h
//
// description : Include file for the AlarmNotify class
//
// project :     Alarm Notify
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef AlarmNotify_H
#define AlarmNotify_H

#include <tango.h>
#include <mutex>
#include <shared_mutex>
#include <unordered_map>

#define NAME_KEY			"tag"
#define RECEIVERS_KEY		"receivers"
#define GROUPS_KEY			"group"


/*----- PROTECTED REGION END -----*/	//	AlarmNotify.h

#ifdef TANGO_LOG
	// cppTango after c934adea (Merge branch 'remove-cout-definition' into 'main', 2022-05-23)
	// nothing to do
#else
	// cppTango 9.3-backports and older
	#define TANGO_LOG       cout
	#define TANGO_LOG_INFO  cout2
	#define TANGO_LOG_DEBUG cout3
#endif // TANGO_LOG

/**
 *  AlarmNotify class description:
 *    Alarm handler notifier
 */


namespace AlarmNotify_ns
{
/*----- PROTECTED REGION ID(AlarmNotify::Additional Class Declarations) ENABLED START -----*/

//	Additional Class Declarations
typedef struct {
	vector<string> alm_list;
	vector<double> freq_list;
	std::shared_ptr<Tango::DeviceProxy> dp;
} handler_t;

bool compare_tango_names(string str1, string str2);

struct tango_names_compare {
  bool operator() (string str1, string str2) const
  {return compare_tango_names(str1,str2);}
};

typedef map<string,handler_t,tango_names_compare> handler_map_t;

typedef struct {
	vector<string> emails;
	//TODO: other receivers
} receivers_t;

//state -> email addr list
typedef map<string, receivers_t> state_addr_map_t;

//alm name -> states-addr map
//groupname -> states-addr map
typedef map<string, state_addr_map_t> receiver_map_t;

//alm name -> groupnames
typedef map<string, string> group_map_t;

//alm name -> state
typedef map<string, string> state_map_t;

struct AlarmStringUtils
{
	static map<string, string> domain_map;
    static map<Tango::DevState,string> tango_states;
	static vector<string> alarm_states;
	static void string_explode(string str, const string& separator, vector<string>& results);
	static void string_vector2map(const vector<string> &str, const string &separator, unordered_map<string,string> &results);
	static void fix_tango_host(string &attr);
	static void add_domain(string &str);
	static string get_only_attr_name(const string &str);
	static string get_only_tango_host(const string &str);
	static void parse_tango_host(const string &th, string &host, int &port);
	static string remove_domain(const string &str);
	static void split_tag_handler(const string &name, string &handler, string &tag);
	static auto compare_without_domain(const string &str1, const string &str2) -> bool;
	static void parse_receivers(const string &receivers_str, state_addr_map_t &st_receivers);
	static void print_receivers(string &receivers_str, const state_addr_map_t &st_receivers);
};


class SubscribeThread;
class PushThread;
class SharedData;

class Receivers :Tango::LogAdapter
{
private:
	ReadersWritersLock      receiverlock;
	receiver_map_t receiverMap;
	group_map_t groupMap;
public:
	Receivers(Tango::DeviceImpl *dev):Tango::LogAdapter(dev){};
	void get_alarm_list_db(const vector<string> &handler_list, vector<string> &al_list);
	receivers_t get_receivers(const string &name, const string &state);
	state_addr_map_t get_state_addr(const string &name);
	string get_groups(const string &name);
	void print_receivers(const string &name, string &out);
};

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::Additional Class Declarations

class AlarmNotify : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(AlarmNotify::Data Members) ENABLED START -----*/

//	Add your own data members
//	static map<string, string> domain_map;
	timespec last_stat;
	vector<string> attribute_search_list_str;
	std::mutex attributes_mutex;
	std::unique_ptr<SubscribeThread, std::function<void(SubscribeThread*)>> thread;
public:
	vector<string> alarm_list;
	vector<string> handler_list_fix;
	//receiver_map_t receiverMap;
	receiver_map_t groupReceiverMap;
	state_addr_map_t globalReceiverMap;
	handler_map_t handlerMap;
	state_map_t prevStateMap;
	std::unique_ptr<PushThread, std::function<void(PushThread*)>> push_thread;
	std::chrono::time_point<std::chrono::system_clock> startup;
	

	uint AttributePendingNumber{0};
	uint AttributeMaxPendingNumber{0};
	uint AttributeStartedNumber{0};
	uint AttributeStoppedNumber{0};
	uint AttributeNumber{0};
	bool subscribe_change{true};
	std::shared_ptr<SharedData> shared;
	//int period{30000}; //TODO: device property
	std::unique_ptr<Receivers, std::function<void(Receivers*)>> alarm_receivers;

	void prepare_receivers(const string &attr_name, const string &state, const string &groups, string &receivers);

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::Data Members

//	Device property data members
public:
	//	HandlerList:	List of alarm handlers to manage
	std::vector<std::string>	handlerList;
	//	PropertyList:	List of free properties where to find alarm handlers to manage
	//  Ex:
	//  AlarmHandler/DeviceList1
	//  AlarmHandler/DeviceList2
	std::vector<std::string>	propertyList;
	//	MaxSearchSize:	Max size of search result
	Tango::DevLong	maxSearchSize;
	//	Configuration:	
	std::vector<std::string>	configuration;
	//	Receivers:	Alarm states and email addresses in the format
	//  Alarm states(comma separated): email addresses (comma separated);
	//  Alarm states 2(comma separated): email addresses 2 (comma separated);
	//  ...
	std::string	receivers;
	//	SubscribePeriod:	Period in seconds
	Tango::DevLong	subscribePeriod;
	//	GroupReceivers:	Each row Group and Alarm states and email addresses in the format
	//  GroupName=Alarm states(comma separated): email addresses (comma separated);
	//  Alarm states 2(comma separated): email addresses 2 (comma separated);
	//  ...
	std::vector<std::string>	groupReceivers;
	//	AlarmFloodPeriod:	Period in seconds to compute statistics for Alarm Flood
	Tango::DevLong	alarmFloodPeriod;
	//	AlarmFloodThresh:	Maximum number of notifications per Alarm per period
	Tango::DevLong	alarmFloodThresh;

//	Attribute data members
public:
	Tango::DevString	*attr_alarmStatus_read;
	Tango::DevString	*attr_handlerStatus_read;
	Tango::DevString	*attr_alarmReceivers_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	AlarmNotify(Tango::DeviceClass *cl,std::string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	AlarmNotify(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	AlarmNotify(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~AlarmNotify();


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();


//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method     : AlarmNotify::read_attr_hardware()
	 *	Description: Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(std::vector<long> &attr_list);

/**
 *	Attribute alarmStatus related methods
 *	Description: List of:
 *               alarm name - status - notification counter
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_alarmStatus(Tango::Attribute &attr);
	virtual bool is_alarmStatus_allowed(Tango::AttReqType type);
/**
 *	Attribute handlerStatus related methods
 *	Description: List of:
 *               Handler Device Name = Tango State
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 1000
 */
	virtual void read_handlerStatus(Tango::Attribute &attr);
	virtual bool is_handlerStatus_allowed(Tango::AttReqType type);
/**
 *	Attribute alarmReceivers related methods
 *	Description: List of:
 *               alarm name - alarm states: receivers; alarm state 2: receivers
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_alarmReceivers(Tango::Attribute &attr);
	virtual bool is_alarmReceivers_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method     : AlarmNotify::add_dynamic_attributes()
	 *	Description: Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command GetAlarmInfo related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarStringArray *get_alarm_info(Tango::DevString argin);
	virtual bool is_GetAlarmInfo_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method     : AlarmNotify::add_dynamic_commands()
	 *	Description: Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(AlarmNotify::Additional Method prototypes) ENABLED START -----*/

//	Additional Method prototypes
	//void parse_receivers(const string &receivers_str, state_addr_map_t &st_receivers);
	void _remove_attribute(size_t idx, bool running, bool stopped);

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::Additional Method prototypes
};

/*----- PROTECTED REGION ID(AlarmNotify::Additional Classes Definitions) ENABLED START -----*/

//	Additional Classes Definitions

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::Additional Classes Definitions

}	//	End of namespace

#endif   //	AlarmNotify_H
